import os
import requests
from requests_toolbelt.multipart import encoder
from clint.textui.progress import Bar as ProgressBar
import json
import tempfile
import shutil
import click

def default_config():
    return {
        'server' : 'https://yadage.cern.ch'
    }

DEFAULT_CONFIGFILE = os.path.expanduser('~/.ydgsvc.json')


def load_config(configfile):
    config = default_config()
    if os.path.exists(configfile):
        config.update(**json.load(open(configfile)))
    return config

def upload_file(filepath,cfg):
    def create_callback(encoder):
        encoder_len = encoder.len
        bar = ProgressBar(expected_size=encoder_len, filled_char='=')

        def callback(monitor):
            bar.show(monitor.bytes_read)

        return callback

    e = encoder.MultipartEncoder(
        fields={'upload_file': ('filename', open(filepath,'rb'), 'text/plain')}
    )
    m = encoder.MultipartEncoderMonitor(e, create_callback(e))

    head = headers(cfg)
    head['Content-Type'] = m.content_type
    r = requests.post('{}/upload'.format(cfg['server']), data=m, verify = False,headers=head)

    response = r.json()
    return response

# Thanks SO
# https://stackoverflow.com/questions/16694907/how-to-download-large-file-in-python-with-requests-py
def download_file(url,stream, request_opts):
    # NOTE the stream=True parameter
    r = requests.get(url, stream=True, **request_opts)
    for chunk in r.iter_content(chunk_size=1024):
        if chunk: # filter out keep-alive new chunks
            stream.write(chunk)

def headers(config):
    auth_token = os.environ.get('YAD_TOKEN')
    if not auth_token:
        auth_token = config.get('auth_token')
    if not auth_token:
        raise RuntimeError('could not find auth token, either set YAD_TOKEN or edit ~/.ydgsvc.json')

    return {
        'Authorization': 'Bearer {}'.format(auth_token),
        'Content-Type': 'application/json'
    }

class Client(object):
    def __init__(self,config = None ):
        self.configfile = config or DEFAULT_CONFIGFILE
        self.config = load_config(self.configfile)

    def submit(self,workflow,toplevel,parameters,input,outputs,local):
        wflowname = 'submit'
        if not outputs:
            raise RuntimeError('need some outputs')
        outputs = ','.join(outputs)

        inputURL = ''
        inputAuth = False
        if input and not input.startswith('http'):
            if os.path.exists(input):
                if os.path.isdir(input):
                    tmpfilename = tempfile.mktemp()
                    shutil.make_archive(tmpfilename, 'zip', input)
                    r = upload_file(tmpfilename+'.zip', self.config)
                    os.remove(tmpfilename+'.zip')
                else:
                    r = upload_file(input, self.config)
                inputURL = '{}/workflow_input/{}'.format(self.config['server'], r['file_id'])
                inputAuth = True
            else:
                raise RuntimeError('not sure how to handle input')
        submit_url = '{}/workflow_submit'.format(self.config['server'])

        submission_data = {
          "outputs": outputs,
          "inputURL": inputURL,
          "inputAuth": inputAuth,
          "preset_pars": parameters,
          "wflowname": wflowname
        }

        if local:
            import yadageschemas
            submission_data['workflow']=yadageschemas.load(workflow, toplevel or os.getcwd(), 'yadage/workflow-schema')
            submission_data['toplevel']=''
        else:
            submission_data['workflow']=workflow
            submission_data['toplevel']=toplevel

        r = requests.post(submit_url,
            data = json.dumps(submission_data),
            headers = headers(self.config),
            verify=False
        )
        if not r.ok:
            raise RuntimeError('submission failed %s', r.content)

        response = r.json()
        return response

    def status(self,workflow_id):
        status_url = '{}/jobstatus/{}'.format(self.config['server'],workflow_id)
        r = requests.get(status_url, verify = False, headers = headers(self.config))
        if not r.ok:
            raise RuntimeError('submission failed %s', r.content)
        return r.json()

    def get(self,workflow_id,resultfile, output, stdout):
        result_url = '{}/results/{}/{}'.format(self.config['server'],workflow_id,resultfile)

        import sys
        if stdout:
            output_stream = sys.stdout
        else:
            output_stream = open(output or result_url.split('/')[-1], 'wb')

        download_file(result_url, output_stream, request_opts = dict(
            verify = False,
            headers = headers(self.config)
        ))
        output_stream.close()

    def job_logs(self,jobid,json_lines,topic):
        for i in requests.get("{}/subjob_logs/{}?topic={}".format(self.config['server'],jobid,topic),
                              headers = headers(self.config), verify=False, stream=True
                              ).iter_lines():
            if not json_lines:
                yield json.loads(i)['msg']
            else:
                yield i

    def wflow_logs(self,wflowid,json_lines,topic):
        if topic == 'state': key = 'state'

        r = requests.get("{}/wflow_logs/{}?topic={}".format(self.config['server'],wflowid,topic), headers = headers(self.config), verify=False, stream=True)

        if topic == 'state':
            for i in r.iter_lines():
                pass
            state = json.loads(i)['state']
            for node in state['dag']['nodes']:
                click.secho('node: {} state: {}'.format(node['name'], node['state']))

        if topic == 'log':
            for i in r.iter_lines():
                if not json_lines:
                    click.secho(json.loads(i)['msg'])
                else:
                    click.secho(i)

    def login(self,server,auth_token):
        if server:
            self.config['server'] = server
        self.config['auth_token'] = auth_token
        json.dump(self.config,open(self.configfile,'w'))

    def upload(self,filepath):
        response = upload_file(filepath,self.config)
        return response



import requests
import click
import os
import json
import yaml
import requests
from .client import Client

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

def options_from_eqdelimstring(opts):
    options = {}
    for x in opts:
        key, value = x.split('=')
        options[key] = yaml.load(value)
    return options

@click.group()
def yad():
    pass

@yad.command()
@click.argument('workflow')
@click.option('-t','--toplevel', help = 'toplevel', default = None)
@click.option('--local/--remote', default = False)
@click.option('-c','--config', help = 'config file', default = None)
@click.option('-p','--parameter', help = 'output', multiple = True)
@click.option('-f','--parameter_file', help = 'output')
@click.option('--json-output/--human-readable', default = False, help = 'JSON output?')
@click.option('-i','--input', help = 'input')
@click.option('-o','--output', help = 'output', multiple = True)
def submit(workflow,toplevel,json_output,local,input,config, output,parameter, parameter_file):
    if parameter_file:
        parameters = yaml.load(open(parameter_file))
    else:
        parameters = {}
    parameters.update(**options_from_eqdelimstring(parameter))

    c = Client(config)
    response = c.submit(workflow,toplevel,parameters,input,output,local)

    if not json_output:
        click.secho('submitted workflow. Monitor it at  {}/monitor/{}'.format(c.config['server'],response['jobguid']))
    else:
        click.echo(json.dumps(response))



@yad.command()
@click.argument('workflow_id')
@click.option('-c','--config', help = 'config file', default = None)
def status(config, workflow_id):
    c = Client(config)
    status = c.status(workflow_id)
    click.echo(json.dumps(status))

@yad.command()
@click.argument('workflow_id')
@click.argument('resultfile')
@click.option('-c','--config', help = 'config file', default = None)
@click.option('-o','--output')
@click.option('--stdout/--no-stdout')
def get(config,workflow_id,resultfile, output, stdout):
    c = Client(config)
    c.get(workflow_id,resultfile, output, stdout)

@yad.command()
@click.argument('filepath')
@click.option('-c','--config', help = 'config file', default = None)
def upload(filepath, config):
    c = Client(config)
    response = c.upload(filepath)
    click.secho('\n')
    click.secho('uploaded file id is: {}'.format(response['file_id']))

@yad.command()
@click.option('-s','--server', help = 'server', default = None)
@click.option('-c','--config', help = 'config file', default = None)
def login(server,config):

    c = Client(config)

    click.secho('''
yadage service
--------------
Hi, if you already have a API key for {server}
please enter it below.

If not, please visit {server}/profile (make sure
you are logged in or click 'Login' on the upper right
hand side) and click 'Generate API Key'
'''.format(server = c.config['server']))
    auth_token = click.prompt('Please enter your API key', hide_input = True)

    c.login(server,auth_token)

@yad.command()
@click.argument('jobid')
@click.option('--json-lines/--no-json', default = False)
@click.option('-s','--server', help = 'server', default = None)
@click.option('-t','--topic', help = 'topic', default = 'run')
@click.option('-c','--config', help = 'config file', default = None)
def job_logs(jobid,server,config,json_lines,topic):
    c = Client(config)
    for l in c.job_logs(jobid,json_lines,topic):
        click.secho(l)

@yad.command()
@click.argument('wflowid')
@click.option('--json-lines/--no-json', default = False)
@click.option('-s','--server', help = 'server', default = None)
@click.option('-t','--topic', help = 'topic', default = 'log')
@click.option('-c','--config', help = 'config file', default = None)
def wflow_logs(wflowid,server,config,json_lines,topic):
    c = Client(config)
    c.wflow_logs(wflowid,json_lines,topic)



from setuptools import setup, find_packages

setup(
  name = 'yadage-service-cli',
  version = '0.1.9',
  description = 'yadage service command line tools',
  url = 'http://github.com/yadage/yadage-service-cli',
  author = 'Kyle Cranmer, Lukas Heinrich',
  author_email = 'cranmer@cern.ch, lukas.heinrich@cern.ch',
  packages = find_packages(),
  entry_points = {
    'console_scripts': [
      'yad = yadagesvccli.cli:yad',
    ]
  },
  install_requires = [
      'click',
      'requests',
      'pyyaml',
      'requests_toolbelt',
      'clint'
  ],
  extras_require = {
    'local' : [
       'yadage-schemas'
    ]
  }
)
